
#ifndef _CAR_H
  #define _CAR_H

  #ifdef __cplusplus
  extern "C" {
  #endif

  #include <stdio.h>

  #define CAR_HEADERSIZE 512
  #define CAR_DESCSIZE 64
  #define CAR_LABELSIZE 192
  #define CAR_DESC "Cybertek ARchive"
  #define CAR_VERSION 0x00010000
  #define CAR_ID 0x52414343

  #define CAR_EMULATION NULL

  typedef struct
  {
    char desc [CAR_DESCSIZE];
    char label [CAR_LABELSIZE];
    long id;
    long version;
    long filesCount;
    char reserved [CAR_HEADERSIZE -
		     CAR_DESCSIZE -
		     CAR_LABELSIZE -
		     sizeof (long) -
		     sizeof (long) -
		     sizeof (long)];
  } TCAR_header;

  #define CAR_FCACHE 1
  #define CAR_FLZW 2

  #define CAR_ITEMSIZE 64

  typedef struct
  {
    long type;
    char name [8+4];
    long date;
    long time;
    long attrib;
    long len;
    long packedLen;
    long crc;
    long begin;
    char *buf;
    char reserved [CAR_ITEMSIZE - (8+4) - 9 * sizeof (long)];
  } TCAR_item;

  #define CAR_TEMU 1

  typedef struct
  {
    long type;
    TCAR_header header;
    TCAR_item *item;
    long curItem;
    long curPos;
    FILE *file;
    long fPos;
  } TCAR;

  extern TCAR *CAR_open (const char *FileName);// = CAR_EMULATION
  /* opens CAR file.

     :FileName = DOS specification ("COMMAND.COM")

     On success returns CAR handle
     On error returns NULL -> invalid handle or CAR file not found.
  */

  extern long CAR_seek (TCAR *car, const char *FileName, long offset); // = 0
  /*
     Scans, CAR file for FileName, sets acual file position and
     returns size of data file stored in CAR file.

     :CAR_Handle = CAR handle. For a new handle call CAR_Open
     :FileName = DOS specification ("COMMAND.COM")

     On success returns size of stored file

     On error returns -1 -> data file not found or invalid Handle
  */


  extern int CAR_read (TCAR *car, void *buf, unsigned len);
  /* reads data from CAR file
     use CAR_select before this function !.

     :Buf = buffer to place data in
     :Length = count of bytes to read (0 - 65534)

     On success returns a positive integer indicating the number of
       bytes placed in the buffer.

     On error returns -1 -> data file not found or invalid handle
  */
  extern int CAR_getc (TCAR *car);

  extern void *CAR_load (TCAR *car, const char *fileName);

  extern void CAR_close (TCAR *car);
  /*
    Closes CAR file, frees ocupied memory and clears used Handle

     :CAR_Handle = CAR handle. For a new handle call CAR_Open

  */

  extern void CAR_loadCache (TCAR *car, const char *fileName);
  extern void CAR_unloadCache (TCAR *car, const char *fileName);
  extern void CAR_loadCacheAll (TCAR *car);
  extern void CAR_unloadCacheAll (TCAR *car);

  extern int CAR_create (const char *CARFileName, const char *files);

  extern int CAR_extract (const char *CARFileName, const char *file);

  #ifdef __cplusplus
  }
  #endif

#endif
