
#include <mem.h>
#include <stdio.h>
#include <dos.h>
#include <stdlib.h>
#include <string.h>
#include <conio.h>

//#include "vga.h"
#include "vesa.h"
#include "svga.h"
#include "dpmi.h"
#include "scroll.h"
#include "pal.h"

#include "keyb.h"

#include "b256_.cpp"

#define VESA_CALL 0x4F
#define VESA_GETINFOBLOCK 0x00
#define VESA_GETMODEINFO 0x01
#define VESA_SWITCHVIDEOMODE 0x02
#define VESA_RETURNVIDEOMODE 0x03
#define VESA_SAVERESTORESTATE 0x04
  #define STATE_GETBUFFERSIZE 0x00
  #define STATE_SAVE 0x01
  #define STATE_RESTORE 0x02
  #define STATE_MASK_ALL 7// 0xF
#define VESA_DAC 0x08
  #define DAC_SETLEN 0x00
  #define DAC_GETLEN 0x01
#define VESA_INITPMINFO 0x0A
#define VESA_SCANLINELEN 0x06
  #define SCANLINE_SETPIXELS 0x00
  #define SCANLINE_GETPIXELS 0x01
  #define SCANLINE_SETBYTES 0x02
  #define SCANLINE_GETBYTES 0x03
#define VESA_SETBANK 0x05
  #define SETBANK_WINDOWA 0x00
  #define SETBANK_WINDOWB 0x01
#define VESA_SETDISPLAYSTART 0x07
  #define VRT_WAIT 0x80
  #define VRT_NOWAIT 0x00
  #define DISPLAYSTART_SET 0x00
  #define DISPLAYSTART_GET 0x01

#define S3_OEMSTRING "S3 Incorporated"
#define MATROX_OEMSTRING "Matrox"

#define VESA_BLOCKLEN 512

#define BLOCK_64K (65536L)
#define VESA_MAXVIDEOMEM ((8192*1024)-1)

//  #define DOUBLE_SCANLINE_PLUS 30

#define MODE_LINEAR 0x80
#define LINEAR_FRAME_BUFFER 0x4000

#define DAC_SWITCHABLE 1

#define MODE_SUPPORTED 1
#define MODE_BIOS 2
#define MODE_COLOR 8
#define MODE_GRAPHICS 16

#define WIN_SUPPORTED 1
#define WIN_READABLE 2
#define WIN_WRITEABLE 4

#define MEM_PLANAR 3             // Planar memory model
#define MEM_PACKEDPIXEL 4        // Packed pixel memory model
#define MEM_RGB 6                // Direct color RGB memory model
#define MEM_YUV 7                // Direct color YUV memory model

#define VESA_SUCCESS 0x004F

#define VESA_X_PLUSMAX 32
#define VESA_Y_PLUSMAX 30

static void far *VESA_linearBuf;
int VESA_version;
char *VESA_OEMString;
int *VESA_modes;
long VESA_memory;

static char S3_OEMString [] = S3_OEMSTRING;
static char MATROX_OEMString [] = MATROX_OEMSTRING;

int VESA_S3Mode;
int VESA_S3;
int VESA_matrox;
int VESA_matroxMode;

  struct TVESA_infoBlock {
    char  Signature [4];         // 4 signature bytes
    short version;                  // VESA version numbers
    short OEMStringOff;            // Pointer to OEM string
    short OEMStringSeg;
    long  Capabilities;            // capabilities of the video environment
    short VideoModeOff;          // pointer to supported Super VGA modes
    short VideoModeSeg;
    short TotalMemory;           // Number of 64kb memory blocks on board

    // VBE 2.0 extension information

    short OemSoftwareRev;            // OEM Software revision number
    long  OemVendorNamePtr;   // Pointer to Vendor Name string
    long  OemProductNamePtr;  // Pointer to Product Name string
    long  OemProductRevPtr;   // Pointer to Product Revision str
    char  reserved[222];           // Pad to 256 byte block size
    char  OemDATA[256];      // Scratch pad for OEM data                        */
  };




  struct TVESA_modeInfo {

                                   // mandatory information

    unsigned short ModeAttributes; // mode attributes
    char        WinAAttributes;          // window A attributes
    char        WinBAttributes;          // window B attributes
    short WinGranularity;        // window granularity
    short WinSize;               // window size
    short       WinASegment;             // window A start segment
    short       WinBSegment;             // window B start segment
    long        WinFuncPtr;              // pointer to windor function
    short       BytesPerScanLine;        // bytes per scan line

                                   // formerly optional information (now mandatory)

    short       XResolution;             // horizontal resolution
    short       YResolution;             // vertical resolution
    char        XCharSize;               // character cell width
    char        YCharSize;               // character cell height
    char        NumberOfPlanes;          // number of memory planes
    char        BitsPerPixel;            // bits per pixel
    char        NumberOfBanks;           // number of banks
    char        MemoryModel;             // memory model type
    char        BankSize;                // bank size in kb
    char        NumberOfImagePages;      // number of images
    char        Reserved;                // reserved for page function

                                         // new Direct Color fields

    char        RedMaskSize;             // size of direct color red mask in bits
    char        RedFieldPosition;        // bit position of LSB of red mask
    char        GreenMaskSize;           // size of direct color green mask in bits
    char        GreenFieldPosition;      // bit position of LSB of green mask
    char        BlueMaskSize;            // size of direct color blue mask in bits
    char        BlueFieldPosition;       // bit position of LSB of blue mask
    char        RsvdMaskSize;            // size of direct color reserved mask in bits
    char        RsvdFieldPosotion;
    char        DirectColorModeInfo;     // Direct Color mode attributes

    // VESA 2.0 extension information
    unsigned long PhysBasePtr;           // Physical address for linear buf
    unsigned long OffScreenMemOffset;    // Pointer to start of offscreen mem
    short         OffScreenMemSize;      // Amount of offscreen mem in 1K's

    // VBE 2.1 extension information
    short   LinBytesPerScanLine;    // Bytes per scanline
    char    BnkNumberOfImagePages;  // Number of images pages (banked)
    char    LinNumberOfImagePages;  // Number of images pages (linear)
    char    LinRedMaskSize;         // Size of direct color red mask
    char    LinRedFieldPosition;    // Bit posn of lsb of red mask
    char    LinGreenMaskSize;       // Size of direct color green mask
    char    LinGreenFieldPosition;  // Bit posn of lsb of green mask
    char    LinBlueMaskSize;        // Size of direct color blue mask
    char    LinBlueFieldPosition;   // Bit posn of lsb of blue mask
    char    LinRsvdMaskSize;        // Size of direct color res mask
    char    LinRsvdFieldPosition;   // Bit posn of lsb of res mask

    char    res2[194];              // Pad to 256 byte block size
  } ;

static int bankShift; // Bank granularity adjust factor
static struct TVESA_modeInfo VESA_modeInfo;
struct TVESA_infoBlock VESA_infoBlock;
void *VESA_DOSBuf;

//static int VESA_displayStartDiv;

static int MemoryModel, MaxPage;
static long PageSize;
static unsigned char redMask,greenMask,blueMask;
static int redPos,redAdjust,greenPos,greenAdjust,bluePos,blueAdjust;

static void *VESA20_setWindowPtr;
static void *VESA20_setPalettePtr;
static void *VESA20_setDisplayStartPtr;

typedef struct {
        short   setWindow;               // Offset of Set Window call
        short   setDisplayStart;         // Offset of Set Display Start call
        short   setPalette;              // Offset of Set Primary Palette
        short   IOPrivInfo;              // Offset of I/O priveledge info
                                         // ... code and tables located here
        } TVESA_PMInfo;

static TVESA_PMInfo *VESA_PMInfo;



static int VESA_call (union RMINFO *rmi)
{
  rmi->h.ah = VESA_CALL;
  DPMI_int (VGA_INT, rmi);
  if (rmi->w.ax == VESA_SUCCESS)
    return 0;
  else
    return 1;
}


int VESA_switchVideoMode (int videoMode)
{
  REGS regs;

//  SVGA_waitRetrace (1);
//  PAL_clear (0, 256);
//  SVGA_waitRetrace (5);
//  SVGA_switchVideoMode (VGA_640x400x16);

  regs.h.ah = VESA_CALL;
  regs.h.al = VESA_SWITCHVIDEOMODE;
  regs.x.ebx = videoMode & ~DOUBLE_BUFFER;
  int386 (VGA_INT, &regs, &regs);

  if (regs.w.ax == VESA_SUCCESS)
  {
    return 0;
  }
  return -1;
}


int VESA_returnVideoMode (void)
{
  REGS regs;

  regs.h.ah = VESA_CALL;
  regs.h.al = VESA_RETURNVIDEOMODE;
  int386 (VGA_INT, &regs, &regs);
  if (regs.w.ax == VESA_SUCCESS)
  {
    return regs.w.bx;
  }
  return -1;
}

static int VESA_getSaveStateBufferLen (void)
{
  RMINFO rmi = {0};

  rmi.h.al = VESA_SAVERESTORESTATE;
  rmi.h.dl = STATE_GETBUFFERSIZE;
  rmi.w.cx = STATE_MASK_ALL;
  if (VESA_call (&rmi) == 0)
  {
    return 64*rmi.w.bx;
  }
  else
  {
    return 0;
  }
}


void *VESA_saveState (void)
{
  RMINFO rmi = {0};
  void *buf;
  int len;

  if (VESA_version)
  {
    len = VESA_getSaveStateBufferLen ();
    if (len && ((unsigned)len < 64000))
    {
      buf = DPMI_alloc (len);
      if (buf)
      {
        rmi.h.al = VESA_SAVERESTORESTATE;
        rmi.h.dl = STATE_SAVE;
        rmi.w.cx = STATE_MASK_ALL;
        rmi.w.bx = DP_OFF (buf);
        rmi.w.es = DP_SEG (buf);
        if (VESA_call (&rmi) == 0)
        {
          return buf;
        }
        else
          return NULL;
      }
      else
      {
        return NULL;
      }
    }
    else
    {
      return NULL;
    }
  }
  else
    return NULL;
}

void VESA_restoreState (void *stateBuf)
{
  RMINFO rmi = {0};

  if (VESA_version)
  {
    if (stateBuf)
    {
      rmi.h.al = VESA_SAVERESTORESTATE;
      rmi.h.dl = STATE_RESTORE;
      rmi.w.cx = STATE_MASK_ALL;
      rmi.w.bx = DP_OFF (stateBuf);
      rmi.w.es = DP_SEG (stateBuf);
      VESA_call (&rmi);
      DPMI_free (stateBuf);
//      stateBuf = VESA_saveState ();
//      if (stateBuf)
//      {
//        DPMI_DOSFree (stateBuf);
//      }
    }
  }
}


static int VESA_getInfoBlock (TVESA_infoBlock *infoBlock)
{
  RMINFO rmi = {0};

  // VESA function 0 - get VESA info block
  rmi.x.eax = VESA_GETINFOBLOCK;
  rmi.x.es = DP_SEG (VESA_DOSBuf);
  rmi.x.edi = DP_OFF (VESA_DOSBuf);
  if (VESA_call (&rmi) == 0)
  {
    if ((sizeof (TVESA_infoBlock) == VESA_BLOCKLEN))
    {
      memcpy (infoBlock, VESA_DOSBuf, VESA_BLOCKLEN);
      return 0;
    }
  }
  return 1;
}


static int VESA_copyModeInfo (int videoMode, TVESA_modeInfo *modeInfo)
{
  RMINFO rmi = {0};

  // VESA function 1 - get VESA mode info
  memset (VESA_DOSBuf, 0, VESA_BLOCKLEN);
  rmi.x.eax = VESA_GETMODEINFO;
  rmi.x.ecx = videoMode;
  rmi.x.es = DP_SEG (VESA_DOSBuf);
  rmi.x.edi = DP_OFF (VESA_DOSBuf);
  if (VESA_call (&rmi) == 0)
  {
    memcpy (modeInfo, VESA_DOSBuf, VESA_BLOCKLEN);
    return 0;
  }
  return 1;
}


static int VESA_initPMInfo (void)
/****************************************************************************
*
* Function:             InitPMCode      - 32 bit protected mode version
*
* Description:  Finds the address of and relocates the protected mode
*                               code block from the VBE 2.0 into a local memory block. The
*                               memory block is allocated with _alloc() and must be freed
*                               with VESA_freePMCode() after graphics processing is complete.
*
*                               Note that this buffer _must_ be recopied after each mode set,
*                               as the routines will change depending on the underlying
*                               video mode.
*
****************************************************************************/
{
  RMINFO rmi = {0};

  if (VESA_PMInfo == NULL)
  {
//    if (VESA_version >= VESA_20)
    {
      rmi.h.al = VESA_INITPMINFO;
      rmi.x.ebx = 0;
      if (VESA_call (&rmi) == 0)
      {
        if ((VESA_PMInfo = (TVESA_PMInfo *) _alloc(rmi.w.cx)) != NULL)
        {
          /* Relocate the block into our local data segment */
          memcpy (VESA_PMInfo, MK_DP (rmi.x.es,rmi.w.di), rmi.w.cx);
          // PM_mapRealPointer(&sel,&off,sregs.es,regs.x.di);
          // PM_memcpynf(PMInfo,sel,off,regs.x.cx);
          return 0;
        }
      }
    }
  }
  return -1;
}

static void VESA_freePMInfo(void)
/****************************************************************************
/*
/* Function:             VESA_freePMCode
/*
/* Description:  This routine frees the protected mode code blocks that
/*                               we copied from the VBE 2.0 interface. This routine must
/*                               be after you have finished graphics processing to _free up
/*                               the memory occupied by the routines. This is necessary
/*                               because the PM info memory block must be re-copied after
/*                               every video mode set from the VBE 2.0 implementation.
/*
/****************************************************************************/
{
  if (VESA_PMInfo)
  {
    _free(VESA_PMInfo);
    VESA_PMInfo = NULL;
  }
}


static int VESA_set8BitPAL (void)
/****************************************************************************
*
* Function:             set8BitPalette
* Returns:              True if 8 bit wide palette has been set.
*
* Description:  Attempts to set the system into the 8 bit wide palette
*                               mode if supported by the VBE. Returns true on success, false
*                               otherwise.
*
****************************************************************************/
{
  RMINFO rmi = {0};

  rmi.h.al = VESA_DAC;
  rmi.h.bl = DAC_SETLEN;
  rmi.h.bh = 8;
  if (VESA_call (&rmi) == 0)
  {
/*    if (rmi.h.bh != 6)
    {
      rmi.h.al = VESA_DAC;
      rmi.h.bl = DAC_GETLEN;
      rmi.h.bh = 0;
      if (VESA_call (&rmi) == 0)
      {
        if (rmi.h.bh == 8)
        {
          SVGA_palType = PAL_8BIT;
          return 0;
        }
      }
    } */
    if (rmi.h.bh == 8)
    {
      SVGA_palType = PAL_8BIT;
      return 0;
    }
  }
  return 1;
}

static int VESA_set6BitPAL(void)
/****************************************************************************
*
* Function:             set6BitPalette
* Returns:              True if 6 bit wide palette has been set.
*
* Description:  Attempts to set the system back into the 6 bit wide palette
*                               mode if supported by the VBE. Returns true on success, false
*                               otherwise.
*
****************************************************************************/
{
  RMINFO rmi = {0};

  rmi.h.al = VESA_DAC;
  rmi.h.bl = DAC_SETLEN;
  rmi.h.bh = 6;
  if (VESA_call (&rmi) == 0)
  {
    if (rmi.h.bh == 6)
    {
      SVGA_palType = PAL_6BIT;
      return 0;
    }
  }
  return 1;
}

static int VESA_setPixelsPerLine(int pixelsPerLine,int *newBytes, int *newPixels, int *newScanlines)
/****************************************************************************
*
* Function:             VESA_setPixelsPerLine
* Parameters:   pixelsPerLine   - Pixels per scanline
*                               newBytes                - Storage for bytes per line value set
*                               newPixels               - Storage for pixels per line value set
*                               maxScanLines    - Storage for maximum number of scanlines
* Returns:      True on success, false on failure
*
* Description:  Sets the scanline length for the video mode to the specified
*                               number of pixels per scanline. If you need more granularity
*                               in TrueColor modes, use the VESA_setBytesPerLine routine
*                               (only valid for VBE 2.0).
*
****************************************************************************/
{
  RMINFO rmi = {0};

  rmi.h.al = VESA_SCANLINELEN;
  rmi.x.ebx = SCANLINE_SETPIXELS;
  rmi.x.ecx = pixelsPerLine;

  if (VESA_call (&rmi) == 0)
  {
    if (newBytes != NULL)
      *newBytes = rmi.x.ebx;
    if (newPixels != NULL)
      *newPixels = rmi.x.ecx;
    if (newScanlines != NULL)
      *newScanlines = rmi.x.edx;
    return 0;
  }
  else
  {
    return -1;
  }
}

static int VESA_setBytesPerLine(int bytesPerLine, int *newBytes,
                         int *newPixels,int *newScanlines)
/****************************************************************************
*
* Function:             VESA_setBytesPerLine
*                       Only supported by VESA 2.0
* Parameters:   pixelsPerLine   - Pixels per scanline
*                               newBytes                - Storage for bytes per line value set
*                               newPixels               - Storage for pixels per line value set
*                               maxScanLines    - Storage for maximum number of scanlines
* Returns:      True on success, false on failure
*
* Description:  Sets the scanline length for the video mode to the specified
*                               number of bytes per scanline (valid for VBE 2.0 only).
*
****************************************************************************/
{
  RMINFO rmi = {0};

  rmi.h.al = VESA_SCANLINELEN;
  rmi.x.ebx = SCANLINE_SETBYTES;
  rmi.x.ecx = bytesPerLine;
  if (VESA_call (&rmi) == 0)
  {
    if (newBytes != NULL)
      *newBytes = rmi.x.ebx;
    if (newPixels != NULL)
      *newPixels = rmi.x.ecx;
    if (newScanlines != NULL)
      *newScanlines = rmi.x.edx;
    return 0;
  }
  else
  {
    return -1;
  }
}


/////////////////////////////////////////////////////////////////////////////

static int VESA_curBank;

#ifndef __TURBOC__

void VESA20_setWindow (int bank);
  #pragma aux VESA20_setWindow = \
  "xor ebx,ebx"\
  "call [VESA20_setWindowPtr]"\
  parm [edx]\
  modify [eax ebx ecx edx esi edi]

#endif

//long VESA_setBankCount;

void VESA_setBank (int bank)
{
  REGS regs;
  int _bank;

  if (VESA_curBank != bank)
  {
    SVGA_setBankCount++;

    VESA_curBank = bank;                // Save current write bank number
    _bank = bank << bankShift; // Adjust to VESA granularity

    #ifndef __TURBOC__
    if (VESA20_setWindowPtr == NULL)
    #endif
    {
      regs.h.ah = VESA_CALL;
      regs.h.al = VESA_SETBANK;
      regs.x.ebx = SETBANK_WINDOWA;                 // Select Win A
      regs.x.edx = _bank;
      int386 (VGA_INT, &regs, &regs);

//      union REGPACK regp = {0};
//      regp.w.ax = (VESA_CALL << 8) + VESA_SETBANK;
//      regp.w.bx = SETBANK_WINDOWA;
//    regp.w.dx = _bank;

//      intr (VGA_INT, &regp);

/*      union RMINFO rmi = {0};

      rmi.w.eax = (VESA_CALL << 8) + VESA_SETBANK;
      rmi.w.ebx = SETBANK_WINDOWA;
      rmi.w.edx = _bank;

      DPMI_int (VGA_INT, &rmi); */

/*      RMREGS regs;

      regs.h.ah = VESA_CALL;
      regs.h.al = VESA_SETBANK;
      regs.x.bx = SETBANK_WINDOWA;                 // Select Win A
      regs.x.dx = _bank;
      PM_int86 (VGA_INT, &regs, &regs); */

    }

    #ifndef __TURBOC__
    else
    {
      VESA20_setWindow (_bank);
    }
    #endif
  }
}

#ifndef __TURBOC__

void VESA20_setDisplayStart (int x, int y, int wait);
  #pragma aux VESA20_setDisplayStart = \
  "call [VESA20_setDisplayStartPtr]"\
  parm [ecx] [edx] [ebx]\
  modify [eax ebx ecx edx esi edi]

#endif


int VESA_getDisplayStart (int *x,int *y)
/****************************************************************************
*
* Function:     VBE_getDisplayStart
* Parameters:   x,y - Place to store starting address value
* Returns:              True if function was successful.
*
****************************************************************************/
{
  REGS regs = {0};

  regs.x.eax = VESA_CALL;
  regs.h.al = VESA_SETDISPLAYSTART;
  regs.h.bl = DISPLAYSTART_GET;
  regs.h.bh = 0;
  int386 (VGA_INT,&regs,&regs);
  if (regs.w.ax == VESA_SUCCESS)
  {
    *x = regs.x.ecx;
    *y = regs.x.edx;
    return 0;
  }
  else
  {
    return -1;
  }
}

void VESA_setDisplayStart (int x,int y,int waitVRT=VRT_NOWAIT)
/****************************************************************************
*
* Function:     VESA_setDisplayStart
* Parameters:   x,y - Position of the first pixel to display
* Returns:              True if function was successful.
*
* Description:  Sets the new starting display position to implement
*               hardware scrolling.
*
****************************************************************************/
{//aaa
  REGS regs;
  long pos;
  int div;

  if (VESA_S3USE)
  {
    div = SVGA_maxX;
  }
  else
  {
    div = SVGA_virtualX;
  }

  #ifndef __TURBOC__
  if (VESA20_setDisplayStartPtr != NULL)
  {
    pos = (x+y*SVGA_virtualX)/4;
    VESA20_setDisplayStart (lo(pos) , hi(pos), waitVRT);
  }
  else
  #endif
  {
    regs.h.ah = VESA_CALL;
    regs.h.al = VESA_SETDISPLAYSTART;
    regs.h.bl = DISPLAYSTART_SET;
    regs.h.bh = waitVRT ? VRT_WAIT : VRT_NOWAIT;

    pos = x+y*SVGA_virtualX;
    regs.x.ecx = pos % div;
    regs.x.edx = pos / div;
    int386 (VGA_INT, &regs, &regs);
  }
//  if (hi (pos) != VESA_defBank)
  VESA_curBank = -1;
//  VESA_defBank = hi (pos);
}

/////////////////////////////////////////////////////////////////////////////




static void VESA_setDisplayStartDiv (void)
{
  int x,y;

}


void VESA_scroll (void)
{
  if (VESA_MATROXUSE)
  {
    VESA_setDisplayStart (SVGA_curX, SVGA_curY, 0);
  }
  else
  {
    VESA_setDisplayStart (SVGA_curXNext, SVGA_curYNext, 0);
  }
}





static void computePageInfo(TVESA_modeInfo *MODEInfo,int *MaxPage,
        long *PageSize)
/****************************************************************************
*
* Function:     computePageInfo
* Parameters:   modeInfo    - Pointer to valid mode information block
*               maxpage     - Number of display pages - 1
*               pagesize    - Size of each logical display page in bytes
*
* Description:  Computes the number of image pages and size of each image
*               page for a specified video mode.
*
****************************************************************************/
{
    long    memsize,size;

        if (MODEInfo->MemoryModel == MEM_PLANAR)
          memsize = 256 * 1024L;
        else
          memsize = 64 * 1024L;

/*      if extendedflipping
          memsize = memory * 64 * 1024L; */

        size = (long)MODEInfo->BytesPerScanLine * MODEInfo->YResolution;
        if (MODEInfo->BitsPerPixel == 4) {

        /* We have a 16 color video mode, so round up the page size to
         * 8k, 16k, 32k or 64k boundaries depending on how large it is.
         */

        size = (size + 0x1FFFL) & 0xFFFFE000L;
        if (size != 0x2000) {
            size = (size + 0x3FFFL) & 0xFFFFC000L;
            if (size != 0x4000) {
                size = (size + 0x7FFFL) & 0xFFFF8000L;
                                if (size != 0x8000)
                    size = (size + 0xFFFFL) & 0xFFFF0000L;
                }
            }
        }
    else
        size = (size + 0xFFFFL) & 0xFFFF0000L;

        if (MODEInfo->MemoryModel == MEM_PLANAR)
        memsize /= 4;
    if (size <= memsize)
    {
      if (size)
      {
        *MaxPage = (memsize / size) - 1;
      }
    }
    else
        *MaxPage = 0;
    *PageSize = size;
}


static void analyzeModeInfo (void)
{
  /* Emulate RGB modes using a 3 3 2 palette arrangement by default */
  redMask = 0x7;        redPos = 5;     redAdjust = 5;
  greenMask = 0x7;      greenPos = 2;   greenAdjust = 5;
  blueMask = 0x3;       bluePos = 0;    blueAdjust = 6;

  if (VESA_modeInfo.MemoryModel == MEM_PACKEDPIXEL &&
              VESA_modeInfo.BitsPerPixel > 8)
  {
    /* Support old style definitions, which some BIOS'es still use :-( */
    VESA_modeInfo.MemoryModel = MEM_RGB;
    switch (VESA_modeInfo.BitsPerPixel)
    {
      case 15:
        redMask = 0x1F;         redPos = 10;    redAdjust = 3;
        greenMask = 0x1F;       greenPos = 5;   greenAdjust = 3;
        blueMask = 0x1F;        bluePos = 0;    blueAdjust = 3;
        break;
      case 16:
        redMask = 0x1F; redPos = 11;    redAdjust = 3;
        greenMask = 0x3F;       greenPos = 5;   greenAdjust = 2;
        blueMask = 0x1F;        bluePos = 0;    blueAdjust = 3;
        break;
      case 24:
      case 32:
        redMask = 0xFF; redPos = 16;    redAdjust = 0;
        greenMask = 0xFF;       greenPos = 8;   greenAdjust = 0;
        blueMask = 0xFF;        bluePos = 0;    blueAdjust = 0;
        break;
    }
  }
  else
    if (VESA_modeInfo.MemoryModel == MEM_RGB)
    {
      /* Convert the 32k direct color modes of VBE 1.2+ BIOSes to
      * be recognised as 15 bits per pixel modes. */

      if (VESA_modeInfo.BitsPerPixel == 16 && VESA_modeInfo.RsvdMaskSize == 1)
        VESA_modeInfo.BitsPerPixel = 15;

      /* Save direct color info mask positions etc */

      redMask = (0xFF >> (redAdjust = 8 - VESA_modeInfo.RedMaskSize));
      redPos = VESA_modeInfo.RedFieldPosition;
      greenMask = (0xFF >> (greenAdjust = 8 - VESA_modeInfo.GreenMaskSize));
      greenPos = VESA_modeInfo.GreenFieldPosition;
      blueMask = (0xFF >> (blueAdjust = 8 - VESA_modeInfo.BlueMaskSize));
      bluePos = VESA_modeInfo.BlueFieldPosition;
    }
}



int VESA_getModeInfo (int videoMode, int *xRes, int *yRes,
                      int *bytesPerLine,int *bitsPerPixel)
/****************************************************************************
*
* Function:     VESA_GetVESA_modeInfo
* Parameters:   Mode            - Mode to get information about
*               XRes            - Place to store x resolution
*               YRes            - Place to store y resolution
*               bytesPerLine    - Bytes per scanline
*               BitsPerPixel    - Place to store bits per pixel (2^n colors)
*               MemoryModel     - Memory model for mode (planar, packed etc)
*               MaxPage         - Number of display pages - 1
*               PageSize        - Size of each logical display page in bytes
* Returns:      Mode if mode number was valid, -1 if not.
*
* Description:  Obtains information about a specific video mode from the
*               VESA. You should use this function to find the video mode
*               you wish to set, as the new VBE 2.0 mode numbers may be
*               completely arbitrary.
*
****************************************************************************/

{
  int i;
  int mode;

  if (VESA_version)
  {
    mode = (videoMode & ~LINEAR_FRAME_BUFFER & ~DOUBLE_BUFFER);

    if (mode >= VESA_MODE)
    {
      // check if desired mode is available
      for (i = 0; VESA_modes [i] != -1 && VESA_modes [i] != mode; i++)
        ;

      if (VESA_modes [i] != -1)
      {
        if (VESA_copyModeInfo (VESA_modes [i], &VESA_modeInfo) == 0)
        {
          // check if the mode is hardware supported
          if ( (VESA_modeInfo.ModeAttributes & MODE_SUPPORTED) &&
          // only 8-bit pixel supported in this version
            (VESA_modeInfo.BitsPerPixel == 8))
          {
            if (((videoMode & LINEAR_FRAME_BUFFER) == 0) || (VESA_modeInfo.ModeAttributes & MODE_LINEAR))
            {
              *xRes = VESA_modeInfo.XResolution;
              *yRes = VESA_modeInfo.YResolution;
              *bytesPerLine = VESA_modeInfo.BytesPerScanLine;
              *bitsPerPixel = VESA_modeInfo.BitsPerPixel;
              return 0;
            }
          }
        }
      }
    }
  }
  return -1;
}


int VESA_init (void)
{
  int mode;
  int i;
  short *vesaModes;

  if ((VGA_installed) && (SVGA_enableVESA))
  {
    VESA_DOSBuf = (void *) DPMI_alloc (4*VESA_BLOCKLEN);

    if (VESA_DOSBuf != NULL)
    {
      if (VESA_getInfoBlock (&VESA_infoBlock) != -1)
      {
        if (memcmp (VESA_infoBlock.Signature, "VESA", 4*sizeof (char)) == 0)
        {
          VESA_OEMString = (char *) MK_DP (VESA_infoBlock.OEMStringSeg, VESA_infoBlock.OEMStringOff);

          i = 0;
          vesaModes = (short *) MK_DP (VESA_infoBlock.VideoModeSeg,VESA_infoBlock.VideoModeOff);
          if (VESA_modes == NULL)
          {
            VESA_modes = (int * ) _calloc (1024, sizeof (int), NULL);
          }
          else
          {
            memset (VESA_modes, 0, 1024*sizeof (int));
          }
          if (VESA_modes == NULL)
          {
            VESA_version = VESA_NONE;
            return NOT_COMPATIBLE;
          }
          do
          {
            mode = vesaModes [i];
            VESA_modes [i] = mode;
            i++;
          }
          while (mode != -1);

          VESA_memory = (long)VESA_infoBlock.TotalMemory*BLOCK_64K;

          // Check for bogus BIOSes that return a VBE version number that is
          // not correct, and fix it up. Currently the only defined standard
          // is VBE 2.0, and it is doubtful if VBE 2.1 will ever be done (VBE/AF
          // is the new standard). Hence anything above 2.0 is considered not
          // really VBE 2.0 at all. We also check the OemVendorNamePtr for a
          // valid value, and if it is invalid then we also reset to VBE 1.2.

          VESA_version = VESA_infoBlock.version;
          if (VESA_version > VESA_20 && VESA_infoBlock.OemVendorNamePtr == 0)
          {
            VESA_version = VESA_12;
          }

          if (VESA_version < VESA_12)
          {
            VESA_version = VESA_NONE;
            return NOT_COMPATIBLE;
          }

          VESA_S3 = (strstr (VESA_OEMString, S3_OEMString) != NULL)
                    && (VESA_version < VESA_20);

          VESA_matrox = (strstr (VESA_OEMString, MATROX_OEMString) != NULL);

          return VESA_COMPATIBLE;
        }
      }
      DPMI_free (VESA_DOSBuf);
      VESA_DOSBuf = NULL;
    }
  }
  VESA_version = VESA_NONE;
  return NOT_COMPATIBLE;
}

void VESA_done (void)
{
  VESA_freePMInfo ();

  VESA20_setWindowPtr       = NULL;
  VESA20_setPalettePtr      = NULL;
  VESA20_setDisplayStartPtr = NULL;

  if (VESA_modes)
  {
    _free (VESA_modes);
    VESA_modes = NULL;
  }

  if (VESA_DOSBuf != NULL)
  {
    DPMI_free (VESA_DOSBuf);
    VESA_DOSBuf = NULL;
  }

  VESA_S3 = 0;
  VESA_matrox = 0;
  VESA_version = 0;
  VESA_OEMString = NULL;
}


int VESA_setVideoMode (int videoMode)
/****************************************************************************
*
* Function:     VESA_SetVideoMode
* Parameters:   mode    - SuperVGA video mode to set.
* Returns:      True if the mode was set, -1 if not.
*
* Description:  Attempts to set the specified video mode. This routine
*               assumes that the library and SuperVGA have been initialised
*               with the VESA_Init () routine first.
*
****************************************************************************/

{
  union REGS regs;
  int bytes, bits;
  int maxX, maxY;
  int i;
  int mode;

  mode = videoMode & 0x4FFF;

  if (VESA_getModeInfo (videoMode,&maxX,&maxY,
                        &bytes,&bits) != 0)
  {
    return SVGA_SETMODE_NOSUPPORT;
  }
  else
  {
    VGA_switchVideoMode (VGA_640x480x16);

    SVGA_graph = 0;

    SVGA_maxX = maxX;
    SVGA_maxY = maxY;
    SVGA_bitsPerPixel = bits;
    SVGA_bytesPerPixel = bytes/SVGA_maxX;
    SVGA_linearBuf = NULL;
    SVGA_linearOff = NULL;
    SVGA_linearOff2 = NULL;
    SVGA_linearSeg = 0;
    SVGA_videoBuf = MK_DP (VGA_SEG, 0);

    #ifndef __TURBOC__
    if ((VESA_modeInfo.ModeAttributes & MODE_LINEAR) == MODE_LINEAR)
    {
      videoMode |= LINEAR_FRAME_BUFFER;
      mode |= LINEAR_FRAME_BUFFER;
    }
    #endif

    VESA_curBank = -1;

    if (VESA_switchVideoMode (mode) != 0)
    {
      return VESA_SETMODE_SWITCHERROR;
    }

    if (videoMode & LINEAR_FRAME_BUFFER)
    {
      if (VESA_linearBuf == NULL)
      {
        VESA_linearBuf = MK_FP (DPMI_newSelector (VESA_modeInfo.PhysBasePtr, VESA_MAXVIDEOMEM), 0);
      }
      SVGA_linearBuf = VESA_linearBuf;
      SVGA_linearSeg = FP_SEG (SVGA_linearBuf);
      SVGA_linearOff = (void *) FP_OFF (SVGA_linearBuf);
    }



    SVGA_doubleBuffer = ((videoMode & DOUBLE_BUFFER) != 0);
    if (SVGA_doubleBuffer)
    {
      for (i = SVGA_maxX+VESA_X_PLUSMAX; i <= VESA_memory/(2*(SVGA_maxY+VESA_Y_PLUSMAX));  i++)
      {
        if (VESA_setPixelsPerLine (i, &SVGA_virtualX, NULL, &SVGA_scanLines) == 0)
        {
          if (SVGA_scanLines >= 2*(SVGA_maxY+VESA_X_PLUSMAX))
          {
            SVGA_scanLines = 2*(SVGA_maxY+VESA_Y_PLUSMAX);
          }
          if (SVGA_virtualX == i)
          {
            SVGA_virtualY = SVGA_scanLines/2;
            break;
          }
        }
        SVGA_virtualX = -1;
      }

      if (i != SVGA_virtualX)
      {
        SVGA_linearBuf = NULL;
        SVGA_linearOff = NULL;
        SVGA_linearOff2 = NULL;
        SVGA_linearSeg = 0;
        SVGA_switchDefaultMode ();
        return VESA_SETMODE_NOVIRTUAL;
//        CGDK_abort ("Unable to set the specified VESA video mode.\n"
//                    "Your VESA video bios extension is not capable to set the proper scan line length.\n"
//                    "Try to install a VESA 2.0 compliant video driver.\n\n");
      }
    }
    else
    {
      SVGA_virtualX = SVGA_maxX;
      SVGA_virtualY = SVGA_maxY;
      SVGA_scanLines = SVGA_virtualY;
    }
    SVGA_maxXPlus = SVGA_virtualX - SVGA_maxX;
    SVGA_maxYPlus = SVGA_virtualY - SVGA_maxY;

    SVGA_virtual = (long) SVGA_virtualX*SVGA_virtualY;
    SVGA_linearOff2 = (char *) SVGA_linearOff + SVGA_virtual;
    SVGA_max = SVGA_maxX*SVGA_maxY;
    SVGA_curX = SVGA_curY = 0;
//    VESA_setDisplayStartDiv ();
    VESA_scroll ();

    bankShift = 0;
    while ((64 >> bankShift) != VESA_modeInfo.WinGranularity)
      bankShift++;

    analyzeModeInfo ();
    computePageInfo(&VESA_modeInfo,&MaxPage,&PageSize);


//    if (SVGA_bitsPerPixel == 8)
    {
      SVGA_palType = PAL_6BIT;
      VESA_set8BitPAL ();
    }

    #ifndef __TURBOC__
      VESA_freePMInfo ();
      if (VESA_initPMInfo () == 0)
      {
        VESA20_setWindowPtr       = (char*)VESA_PMInfo+VESA_PMInfo->setWindow;
        VESA20_setPalettePtr      = (char*)VESA_PMInfo+VESA_PMInfo->setPalette;
        VESA20_setDisplayStartPtr = (char*)VESA_PMInfo+VESA_PMInfo->setDisplayStart;
      }
      else
      {
        VESA20_setWindowPtr       = NULL;
        VESA20_setPalettePtr      = NULL;
        VESA20_setDisplayStartPtr = NULL;
      }
    #endif

    SVGA_graph = 1;

    return SVGA_SETMODE_OK;
  }
}


int VESA_videoModeSupported (int videoMode)
{
  int xRes, yRes, bytes, bits;
  long mem;

  if ((videoMode & 0xFFF) >= VESA_MODE)
  {
    if (VESA_getModeInfo (videoMode, &xRes, &yRes, &bytes, &bits) == 0)
    {
      mem = (long) (xRes+VESA_X_PLUSMAX)*(yRes+VESA_Y_PLUSMAX);
      if (videoMode & DOUBLE_BUFFER)
      {
        mem = 2*mem;
      }
      return (VESA_memory <= mem);
    }
  }
  return 1;
}
